package minishell;
import java.util.*;
import java.io.*;

/**
 * a csh-like toy shell
 */
public class Shell
{
    public static boolean debug = false;
    static void log(String msg)
    {
	if (!debug)
	    return;
	System.out.println("Shell: " + msg);
    }

    private boolean	interactive = false;

    private Hashtable	variables;
    private Hashtable	aliases;
    private Vector	applications;
    private Properties  env;

    public Shell(boolean interactive)
    {
	this();
	this.interactive = interactive;
    }

    public Shell()
    {
	env = System.getProperties();

	variables = new Hashtable();
	variables.put("prompt", "jsh> ");
	aliases = new Hashtable();
	aliases.put("compile", "pizza.compiler.Main");
	aliases.put("suncompile", "sun.tools.javac.Main");
	applications = new Vector();
	try {
		run(new LineNumberReader(
		    new InputStreamReader(
			new FileInputStream("minishell.rc"))));
	} catch (FileNotFoundException _) { }
    }

    public void run()
    {
	System.out.println("\nWelcome to Java on the OSKit\n");
	System.out.println("The following csh commands are supported:\n");
	System.out.println("exit, set, source, alias, unalias, " 
		+ "env, setenv, unsetenv,\n" 
		+ "shutdown, share, private\n");
	run(new LineNumberReader(new InputStreamReader(System.in)));
    }

    void run(LineNumberReader inreader)
    {
        String [] argv;

        for (;;) {
            try {
		if (interactive) {
		    String prompt = (String)variables.get("prompt");
		    if (prompt == null)
			prompt = "> ";
		    System.out.print(prompt);
		    System.out.flush();
		}

                String l = inreader.readLine();
                if (l == null)
		    break;

                LineListTokenizer lt = new LineListTokenizer(
                            new StringReader(l + "\n"));

                argv = lt.list();
                if (argv == null)
                    continue;

		// first, replace and handle aliases
		argv = handleAliases(argv);

		// second, replace and handle variables
		argv = handleVariables(argv);

		// if nothing is left to be done, continue
                if (argv == null)
                    continue;

		// print environment
                if (argv[0].equals("env")) {
		    handleEnv();
		    continue;
		}

		// setenv variable value
                if (argv[0].equals("setenv")) {
		    handleSetenv(argv);
		    continue;
		}

		// unsetenv {var}*
                if (argv[0].equals("unsetenv")) {
		    handleUnsetenv(argv);
		    continue;
		}

		// private
                if (argv[0].equals("private")) {
		    handlePrivate(argv);
		    continue;
		}

		// share
                if (argv[0].equals("share")) {
		    handleShare(argv);
		    continue;
		}

		// gc
                if (argv[0].equals("gc")) {
		    System.gc();
		    continue;
		}

		// set variable = value
                if (argv[0].equals("set")) {
		    handleSet(argv);
		    continue;
		}

		// exit
                if (argv[0].equals("exit"))
                    break;

		while ((argv = handleCommand(argv)) != null)
			;

            } catch (Exception e) {
		System.err.println("fatal: " + e);
		e.printStackTrace(System.err);
                break;
            }
        }
    }

    private String []handleCommand(String argv[])
    {
	// source file1 [file2 file3 ...]
	if (argv[0].equals("source")) {
	    for (int i = 1; i < argv.length; i++) {
		boolean saved_ia = true;
		try {
		    saved_ia = interactive;
		    interactive = false;
		    run(new LineNumberReader(
			    new InputStreamReader(
				new FileInputStream(argv[i]))));
		} catch (FileNotFoundException e) {
		    System.err.println(argv[i] 
			+ ": No such file or directory.");
		} finally {
		    interactive = saved_ia;
		}
	    }
	    return null;
	}

	boolean cmdcomplete = false;
	boolean background = false;
	Vector ar = new Vector();

	// no fancyness here, just read one by one
	try {
	    Application app = null;

	    for (int i = 0; i < argv.length; i++) {

		if (argv[i].equals("&")) {
		    background = true;	// XXX
		} else
		if (argv[i].equals(";")) {
		    cmdcomplete = true;
		} else {
		    // else simply collect args
		    ar.addElement(argv[i]);
		}
		if (i == argv.length - 1)
		    cmdcomplete = true;

		if (cmdcomplete) {
		    String cp = env.getProperty("java.class.path");
		    String []args = new String[ar.size()];
		    ar.copyInto(args);

		    app = new Application(args, cp);

		    // prepare next instream as the other end of pipe if needed
		    cmdcomplete = false;
		    background = false;
		    ar.removeAllElements();
		}
	    }

	    app.run();

	} catch (Exception e) {
	    System.err.println(e);
	    e.printStackTrace(System.err);
	    return null;
	}
	return null;
    }

    /**
     * print environment
     */
    private void handleEnv()
    {
	env.save(System.out, "current environment settings");
	System.out.flush();
    }

    /**
     * unsetenv {var}*
     */
    private void handleUnsetenv(String argv[])
    {
	if (argv.length < 2) {
	    System.err.println("unsetenv: Too few arguments.");
	    return;
	}
	for (int i = 1; i < argv.length; i++)
	    env.remove(argv[i]);
    }

    /**
     *
     */
    private void handlePrivate(String argv[])
    {
	if (argv.length < 2) {
	    PrintWriter p = new PrintWriter(System.out);
	    ClassPathLoader.listPrivateNames(p);
	    p.flush();
	    return;
	}
	for (int i = 1; i < argv.length; i++)
	    ClassPathLoader.addPrivateName(argv[i]);
    }

    /**
     *
     */
    private void handleShare(String argv[])
    {
	if (argv.length < 2) {
	    PrintWriter p = new PrintWriter(System.out);
	    ClassPathLoader.listSharedPrefixes(p);
	    p.flush();
	    return;
	}
	for (int i = 1; i < argv.length; i++)
	    ClassPathLoader.addSharedPrefix(argv[i]);
    }

    /**
     * setenv var value
     */
    private void handleSetenv(String argv[])
    {
	if (argv.length < 2) {
	    handleEnv();
	    return;
	}

	String var = argv[1];
	String value = argv.length > 2 ? argv[2] : "";
	env.put(var, value);
    }

    /**
     * set var = value
     */
    private void handleSet(String argv[])
    {
	if (argv.length == 1) {
	    Enumeration e = variables.keys();
	    for (; e.hasMoreElements();) {
		String var = (String)e.nextElement();
		String value = (String)variables.get(var);
		System.out.println(var + '\t' + value);
	    }
	    return;
	}

	if (argv.length <= 3)
	    return;

	String var = argv[1];
	String value = argv[3];

	if (var.lastIndexOf(".") == -1) {
	    variables.put(var, value);
	    return;
	} else {
	    try {
		log("Trying to set " + var + " to " + value);
		new SetField(var, value);
	    } catch (Exception e) {
		System.err.println("Couldn't set " + var + ":" + e);
	    }
	}
    }

    private void printAlias(String s)
    {
	String a = (String)aliases.get(s);
	System.out.println(s + "\t" + (a == null ? "<null>" : a));
    }

    private void printAliases()
    {
	for (Enumeration e = aliases.keys(); e.hasMoreElements();)
	    printAlias((String)e.nextElement());
    }

    /**
     * handle and replace shell and environment variables in strings 
     */
    private String [] handleVariables(String argv[])
    {
	return argv;
    }

    /**
     * handle and replace aliases in string (replacement is done in-place, 1:1)
     */
    private String [] handleAliases(String argv[])
    {
	if (argv[0].equals("unalias")) {
	    if (argv.length > 1)
		aliases.remove(argv[1]);
	    else
		System.err.println("unalias: Too few arguments.");
	    return null;
	}

	if (argv[0].equals("alias")) {
	    if (argv.length > 2)
		aliases.put(argv[1], argv[2]);
	    else if (argv.length > 1)
		printAlias(argv[1]);
	    else
		printAliases();
	    return null;
	}

	for (int i = 0; i < argv.length; i++) {
	    String s = (String)aliases.get(argv[i]);
	    if (s != null)
		argv[i] = s;
	}
	return argv;
    }

    public static void main(String av[])
    {
	ClassPathLoader.addSharedPrefix("java.");
	ClassPathLoader.addSharedPrefix("kaffe.");
	new Shell(true).run();
    }
}
